package sofie.maxsat;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.database.Database;
import basics.datagraph.DataGraphManager;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.YagoClasses;

/**
 * Class Configuration
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *
 * Holds global data
 *  
 * @author Fabian M. Suchanek
 */
public class Config {

  /** Maximum ID to consider among the pattern occurrences (or -1)*/
  public static int maxPatternOccurrenceId = -1;

  /** Minimum ID to consider among the pattern occurrences (or -1)*/
  public static int minPatternOccurrenceId = -1;

  /** Batch size */
  public static int BATCHSIZE= 20000;

  /** Points to the database*/
  public static Database database;

  /** Relations to learn (as Strings, because they may be undeclared)*/
  protected static Collection<String> relationsToLearn;

  /** The rules that to optimize*/
  public static List<Rule> rules;

  /** Additional constraint to consider only patternoccurrences < MAXNUMPATOC*/
  public static String delimitPOC = null;

  /** Initializes the variables*/
  public static void init(File iniFile) throws Exception {
    Announce.doing("Initializing from",iniFile);    

    // Load the parameters
    Parameters.init(iniFile);
    DataGraphManager.init();
    Hypothesis.setRelations();
    BATCHSIZE=Parameters.getInt("batchsize", BATCHSIZE);
      
    // Load relations that shall be learned
    relationsToLearn = new ArrayList<String>();
    Parameters.ensureParameters("relationsToLearn - a comma-separated list of relations");
    for (String relationName : Parameters.getList("relationsToLearn")) {
      Relation relation = Relations.get(relationName);
      if(D.indexOf(relation,Relations.MEANS,Relations.TYPE,Hypothesis.DISAMBIGUATIONPRIOR,Hypothesis.PATTERNOCCURRENCES)!=-1) {
        Announce.error("Cannot learn relation",relation);
      }
      if (relation == null) {
        Relations.add(relation=new Relation(relationName,YagoClasses.ENTITY,YagoClasses.ENTITY));        
      }
      relationName = relation.name();
      relationsToLearn.add(relationName);
    }
    relationsToLearn.add(Hypothesis.DISAMBIGUATEDAS.toString());

    // Load the database
    database = Parameters.getDatabase();

    // Load the rules
    Parameters.ensureParameters("rules - points to a file containing the rules");
    rules = Rule.loadFrom(Parameters.getFile("rules"));
    Announce.done();
    
  }

  /** TRUE for relations to learn*/
  public static boolean isRelationToLearn(String rel) {
    return (relationsToLearn.contains(rel));
  }

  /** Normalizes a relation to learn*/
  public static String normalizeRelationToLearn(String rel) {
    for (String s : relationsToLearn) {
      if (s.equalsIgnoreCase(rel)) return (s);
    }
    return (null);
  }

  /** TRUE for relations to learn*/
  public static boolean isRelationToLearn(Relation rel) {
    return (relationsToLearn.contains(rel.name()));
  }

  /** Sets patternoccurence limits */
  public static void setPatternOccurences(int from, int to) {
    minPatternOccurrenceId = from;
    maxPatternOccurrenceId = to;
    if (maxPatternOccurrenceId < 0) {
      delimitPOC = null;
    } else {
      delimitPOC = "patternOccurrences.id<" + maxPatternOccurrenceId + " AND patternOccurrences.id>=" + minPatternOccurrenceId;
    }
  }

  /** Closes the database*/
  public static void close() {
    database.close();
  }

  /** Test */
  public static void main(String[] args) throws Exception {
    init(new File("sofie.ini"));
    D.p(rules);
    D.p(relationsToLearn);
  }
}
