package sofie.maxsat;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.database.Database.Inserter;
import javatools.datatypes.FinalSet;
import javatools.filehandlers.CSVFile;
import javatools.parsers.Char;
import basics.datagraph.LiteralYagoClass;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.VirtualRelation;
import basics.datagraph.YagoClasses;

/**
 * Class Hypothesis
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *      
 * This class prepresents a hypothesis with 2 or more arguments for SOFIE.
 * 
 * @author Fabian M. Suchanek
 */
public class Hypothesis implements Comparable<Hypothesis> {

  // ----------------------------------------------------------------------------
  //                    Special YAGO relations and YAGO classes
  // ----------------------------------------------------------------------------

  /** Special class for wordInContext*/
  public static final LiteralYagoClass WORDINCONTEXT = new LiteralYagoClass("yagoWordInContext", YagoClasses.IDENTIFIER, "^.*");

  /** Special Relation: patternOccurrences(patternId,x,y,count) */
  public static final Relation PATTERNOCCURRENCES = new Relation("patternOccurrences", YagoClasses.ENTITY, YagoClasses.ENTITY);

  /** Special Relation: flagSet(flagName) */
  public static final Relation FLAGSET = new Relation("flagSet", null, null);

  /** Special Relation: disambiguationPrior(wordInContext,entity) */
  public static final Relation DISAMBIGUATIONPRIOR = new Relation("disambiguationPrior", WORDINCONTEXT, YagoClasses.ENTITY);

  /** Special Relation: disambiguatedAs(wordInContext,entity) */
  public static final Relation DISAMBIGUATEDAS = new Relation("disambiguatedAs", WORDINCONTEXT, YagoClasses.ENTITY, YagoClasses.FUNCTION);

  /** DEBUG: If you would like to know what happens to a particular hypothesis,
   * add the hypothesis here!*/
  private static Set<Hypothesis> traceHypotheses = new FinalSet<Hypothesis>(
      //new Hypothesis(true, "bornIn","Fabian_Martin_Suchanek", "Mountain_View,_Hawaii")
      //new Hypothesis(true,"disambiguatedAs","^Mountain View0","Mountain_View,_Hawaii"),
      //new Hypothesis(true,"isGoodPatternFor","1","bornOnDate"),
      //new Hypothesis("bornOnDate","Fabian_Martin_Suchanek", "2009-01-01")
      //new Hypothesis("bornOnDate","Albert_Einstein","1879-03-14")
      //new Hypothesis("disjoint","America_(1911_New_York_automobile)", "America_(2002_novel)"),
      //new Hypothesis("disambiguatedAs", "^America23", "America_(2002_novel)")
      /** Add here*/);

  // ----------------------------------------------------------------------------
  //                    Helper Methods
  // ----------------------------------------------------------------------------

  /** Adds the relations to the gloabl contains*/
  public static void setRelations() {
    Relations.add(DISAMBIGUATIONPRIOR);
    Relations.add(DISAMBIGUATEDAS);
    Relations.add(FLAGSET);
    Relations.add(PATTERNOCCURRENCES);
  }

  /** TRUE for hypotheses that are to be traced*/
  public boolean trace() {
    return(traceHypotheses.contains(this));
  }  

  /** 3-valued Truth values */
  public enum Truth {
    TRUE, FALSE, UNKNOWN;
    public boolean asBoolean() {
      return(this==TRUE);
    }
    public static Truth of(boolean b) {
      return (b ? TRUE : FALSE);
    }

    public static Truth of(double b) {
      return (b < 0 ? UNKNOWN : b < 0.5 ? FALSE : TRUE);
    }
  };

  /** True if a string starts with a '^'*/
  public static boolean isPolysem(String s) {
    return (s != null && s.startsWith("^"));
  }

  /** True if a string is not null and starts not with a '^'*/
  public static boolean isEntity(String s) {
    return (s != null && !s.startsWith("^"));
  }

  /** True if a string is not null and starts with a '"'*/
  public static boolean isString(String s) {
    return (s != null && s.startsWith("\""));
  }

  /** True if a string starts with a '?'*/
  public static boolean isVariable(String s) {
    return (s.startsWith("?"));
  }

  // ----------------------------------------------------------------------------
  //                    Definition of a hypothesis
  // ----------------------------------------------------------------------------

  /** Holds the relation followed by the arguments*/
  protected List<String> components;

  /** Holds the relation, if it is a registered relation*/
  protected Relation relation;

  /** Holds the confidence value*/
  protected double value = -1;

  /** Holds the priority of this hypothesis*/
  protected double priority = -1;

  /** Copies a hypothesis*/
  protected Hypothesis(Hypothesis hyp) {
    this(hyp.value(), hyp.components());
  }

  /** Constructs a fact*/
  public Hypothesis(double value, List<String> components) {
    this.components = components;
    this.relation = components.size() == 0 ? null : Relations.get(components.get(0));
    if (relation != null) components.set(0, relation.toString());
    this.value = value;
  }

  /** Constructs a fact*/
  public Hypothesis(double value, String... components) {
    this(value, Arrays.asList(components));
  }

  /** Constructs a fact*/
  public Hypothesis(double value, Relation rel, String... args) {
    this.components = new ArrayList<String>();
    this.components.add(rel.toString());
    this.components.addAll(Arrays.asList(args));
    this.relation = rel;
    this.value = value;
  }

  /** Constructs a fact*/
  public Hypothesis(boolean value, Relation rel, String... args) {
    this(value ? 1 : 0, rel, args);
  }

  /** Constructs a fact*/
  public Hypothesis(boolean value, String... components) {
    this(value ? 1 : 0, components);
  }

  /** Constructs a fact*/
  public Hypothesis(String... components) {
    this(1, components);
  }

  /** Constructs a fact*/
  public Hypothesis(boolean value, List<String> components) {
    this(value ? 1 : 0, components);
  }

  @Override
  public String toString() {
    StringBuilder b=new StringBuilder(component(0));
    b.append('(');
    for(int i=1;i<numComponents();i++) b.append(component(i)).append(", ");
    if(Char.last(b)==' ') Char.cutLast(Char.cutLast(b));
    b.append(")/").append(value);
    return(b.toString());
  }

  @Override
  public int hashCode() {
    return (components.hashCode());
  }

  @Override
  public boolean equals(Object obj) {
    return (obj != null && obj instanceof Hypothesis && components.equals(((Hypothesis) obj).components));
  }

  /** Returns the priority*/
  public double priority() {
    return priority;
  }

  /** Sets the priority*/
  public void setPriority(double v) {
    priority = v;
  }

  /** Returns the current value*/
  public double value() {
    return (value);
  }

  /** Sets the value*/
  public double setValue(double v) {
    return (value = v);
  }

  /** Sets the value*/
  public double setValue(boolean v) {
    if(trace()) Announce.message("Setting value of",this,"to",v?"TRUE":"FALSE");
    return (value = v ? 1 : 0);
  }

  /** Sets the value*/
  public double setValue(Truth v) {
    switch (v) {
      case TRUE:
        value = 1;
        break;
      case FALSE:
        value = 0;
        break;
      case UNKNOWN:
        value = -1;
        break;
    }
    return (value);
  }

  /** Returns the i-th argument (0-based)*/
  public String arg(int i) {
    return components.get(i + 1);
  }

  /** Returns the i-th component (0-based)*/
  public String component(int i) {
    return components.get(i);
  }

  /** Returns the components */
  public List<String> components() {
    return components;
  }

  /** Returns the relation */
  public Relation relationIfKnown() {
    return relation;
  }

  /** Returns the relation name (or variable)*/
  public String relationName() {
    return components.get(0);
  }

  /** Returns the number of arguments*/
  public int numArgs() {
    return (components.size() - 1);
  }

  /** Returns the number of components*/
  public int numComponents() {
    return (components.size());
  }

  /** TRUE if the relation is virtual*/
  public boolean hasVirtualRelation() {
    return (relation != null && relation instanceof VirtualRelation);
  }

  /** Standard compareTo*/
  public int compareTo(Hypothesis o) {
    // Sort first by number of arguments
    if (o.numComponents()!=numComponents()) return (o.numComponents() > numComponents() ? 1 : -1);
    // Sort by component, sort variables to the begining (important for HypothesisStore)
    for (int i = 0; i < numComponents(); i++) {
      if (Literal.isVariable(component(i)) && !Literal.isVariable(o.component(i))) return (-1);
      if (!Literal.isVariable(component(i)) && Literal.isVariable(o.component(i))) return (1);
      int cmpArg = D.compare(this.component(i), o.component(i));
      if (cmpArg != 0) return (cmpArg);
    }
    return (0);
  }

  /** TRUE is any component is a polysem*/
  public boolean containsPolysem() {
    for (String comp : components()) {
      if (isPolysem(comp)) return (true);
    }
    return (false);
  }

  // ----------------------------------------------------------------------------
  //                    Database interaction
  // ----------------------------------------------------------------------------
  
  /** Stores a hypothesis with 2 arguments to a database*/
  public void saveTo(long id, Inserter hypBulki) throws SQLException {
    hypBulki.insert(id, component(0), component(1), component(2), 0.6);
  }

  /** Stores a hypothesis with 2 arguments to writer*/
  public void saveTo(long id, CSVFile out) throws IOException {
    out.write(id, component(0), component(1) , component(2));
  }

}
