package sofie.maxsat;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.database.ResultIterator;
import javatools.database.ResultIterator.ResultWrapper;
import javatools.datatypes.CombinedIterator;
import javatools.datatypes.PeekIterator;
import javatools.parsers.Char;
import basics.datagraph.LiteralYagoClass;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.VirtualRelation;
import basics.datagraph.YagoClass;
import basics.datagraph.YagoClasses;

/**
 * Class Literal
 * 
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It
 * is licensed under the Creative Commons Attribution-Noncommercial-Share-Alike
 * 3.0 Unported License (http://creativecommons.org/licenses/by-nc-sa/3.0/) by
 * Fabian M. Suchanek (http://suchanek.name). If you use this class for
 * scientific purposes, please cite Fabian M. Suchanek, Mauro Sozio, Gerhard
 * Weikum "SOFIE: A Self-Organizing Framework for Information Extraction"
 * (International World Wide Web Conference 2009)
 * 
 * This class represents a literal, i.e. a hypothesis with a truth value,
 * possibly containing variables.
 * 
 * @author Fabian M. Suchanek
 */
public class Literal extends Hypothesis {

	// ----------------------------------------------------------------------------
	// Definition of a literal
	// ----------------------------------------------------------------------------

	/** Holds the polarity */
	protected boolean polarity = true;

	/** Constructs a literal */
	public Literal(boolean pol, List<String> parts) {
		super(-1, parts);
		polarity = pol;
	}

	/** Constructs a literal */
	public Literal(boolean pol, String... parts) {
		this(pol, Arrays.asList(parts));
	}

	/** Constructs a literal */
	public Literal(boolean pol, Relation r, String... a) {
		super(-1, r, a);
		polarity = pol;
	}

	/** Returns the polarity */
	public boolean polarity() {
		return polarity;
	}

	/** Swaps the polarity, returns THIS */
	public Literal swapPolarity() {
		polarity = !polarity;
		return (this);
	}

	/** Returns the value of f, adjusted for polarity */
	public Truth polarityOf(Hypothesis f) {
		double v = f.value();
		if (v == -1)
			return (Truth.UNKNOWN);
		if (v < 0.001)
			return (polarity() == true ? Truth.FALSE : Truth.TRUE);
		return (polarity() ? Truth.TRUE : Truth.FALSE);
	}

	/** TRUE if the literal is unhappy with this hypothesis */
	public boolean isUnhappyWith(Hypothesis h) {
		if (h.value() == -1)
			return (true);
		return (Truth.of(polarity()) != Truth.of(h.value()));
	}

	@Override
	public String toString() {
		StringBuilder b = new StringBuilder();
		if (polarity() == false)
			b.append('~');
		b.append(component(0));
		b.append('(');
		for (int i = 1; i < numComponents(); i++)
			b.append(component(i)).append(", ");
		if (Char.last(b) == ' ')
			Char.cutLast(Char.cutLast(b));
		b.append(')');
		return (b.toString());
	}

	/**
	 * Returns the bindings of variables of this literal that are necessary to
	 * match this literal to f. NULL for failure
	 */
	public Map<String, String> bindingsToMatch(Hypothesis f,
			Map<String, String> bindings) {
		if (f.numArgs() != numArgs())
			return (null);
		Map<String, String> result = new TreeMap<String, String>();
		if (bindings != null)
			result.putAll(bindings);
		for (int i = 0; i < components.size(); i++) {
			String comp = result.containsKey(component(i)) ? result
					.get(component(i)) : component(i);
			if (isVariable(comp)) {
				result.put(component(i), f.component(i));
				result.put(comp, f.component(i));
			} else {
				if (!comp.equals(f.component(i)))
					return (null);
			}
		}
		return (result);
	}

	/** Applies the bindings */
	public Literal withBindings(Map<String, String> bindings) {
		List<String> comp = new ArrayList<String>(components);
		for (int i = 0; i < comp.size(); i++) {
			if (bindings.containsKey(comp.get(i)))
				comp.set(i, bindings.get(comp.get(i)));
		}
		return (new Literal(polarity, comp));
	}

	/**
	 * Returns the bindings of variables of this literal that are necessary to
	 * match this literal to f. NULL for failure
	 */
	public boolean matches(Hypothesis f) {
		if (f.numArgs() != numArgs())
			return (false);
		for (int i = 0; i < components.size(); i++) {
			if (!isVariable(component(i))
					&& !component(i).equals(f.component(i)))
				return (false);
		}
		return (true);
	}

	/** TRUE for free literals */
	public boolean isFree() {
		return (isVariable(component(0)) || Config.relationsToLearn
				.contains(component(0)));
	}

	/**
	 * Returns how good it would be to ground this literal with the current
	 * known variables
	 */
	public int groundPriority(Collection<String> knownVariables) {
		boolean[] instantiated = new boolean[components.size()];
		int numInstantiated = 0;
		for (int i = 0; i < components.size(); i++) {
			instantiated[i] = !isVariable(component(i))
					|| knownVariables.contains(component(i));
			if (instantiated[i])
				numInstantiated++;
		}
		if (numInstantiated == components.size())
			return (1000);
		if (hasVirtualRelation())
			return (-1000);
		if ((relation == DISAMBIGUATEDAS || relation == DISAMBIGUATIONPRIOR)
				&& instantiated[1])
			return (500);
		if ((relation == DISAMBIGUATEDAS || relation == DISAMBIGUATIONPRIOR)
				&& numInstantiated > 1
				&& !Parameters.getBoolean("POLYSEMS", true))
			return (500);
		int relationPrior = 1;
		if (relation == DISAMBIGUATEDAS)
			relationPrior = 2;
		if (relation == Relations.TYPE)
			relationPrior = 0;
		return (numInstantiated * 10 + relationPrior);
	}

	/** TRUE if the literal contains a variable */
	public boolean hasVariable() {
		for (String s : components) {
			if (isVariable(s))
				return (true);
		}
		return (false);
	}

	// ----------------------------------------------------------------------------
	// Grounding
	// ----------------------------------------------------------------------------

	/**
	 * Helper routine for ground(). Returns the hypothesis only if it makes the
	 * literal unhappy. Sets this's value to value
	 */
	protected Iterable<Hypothesis> singleHypothesis(boolean value) {
		if (trace())
			Announce.message("Return single hypothesis", this, value);
		if (polarity() == value) {
			if (trace())
				Announce.message("Hypothesis does not contradict literal",
						this, ", return EMPTY");
			return (PeekIterator.emptyIterator());
		}
		setValue(value);
		return (Arrays.asList((Hypothesis) this));
	}

	/**
	 * Helper routine for ground(). If the polarity of this literal is TRUE, we
	 * want to return only FALSE groundings. In the closed world assumption,
	 * there are no explicit FALSE groundings. Hence, in this case, we return a
	 * dummy hypothesis with truth value FALSE.
	 */
	protected PeekIterator<Hypothesis> closedWorldIterator(
			PeekIterator<Hypothesis> it) {
		if (trace())
			Announce.message("Hypothesis is closed world", this);
		if (polarity() == true) {
			if (trace()) {
				Announce.message("We want only false hypotheses.");
				if (it.iterator().hasNext())
					Announce.message("However,", this,
							"is actually true. Therefore, return EMPTY");
				else
					Announce.message(this, "is false, return it");
			}
			if (it.iterator().hasNext())
				return (PeekIterator.emptyIterator());
			setValue(false);
			return (new PeekIterator.ElementaryPeekIterator<Hypothesis>(this));
		}
		return (it);
	}

	/**
	 * Returns a set of grounded hypotheses for the literal. Returns only those
	 * hypotheses that make the literal unhappy. Returns a hypothesis with
	 * value=-2 if that hypothesis should be added to the database.
	 */
	public Iterable<Hypothesis> ground() throws SQLException {

		// Mounting
		if (trace())
			Announce.message("Grounding", this);
		if (isVariable(component(0)) && isVariable(component(1))
				&& isVariable(component(2)))
			Announce.error("Grounding general literal", this);

		// Pattern occurrences
		if (this.relationIfKnown() == PATTERNOCCURRENCES) {
			return (closedWorldIterator(this.instances()));
		}

		// Disambiguation prior
		if (this.relationIfKnown() == DISAMBIGUATIONPRIOR) {
			return (closedWorldIterator(this.instances()));
		}

		// Virtual Relations
		if (this.hasVirtualRelation()) {
			if (this.component(1) == null || this.component(2) == null)
				throw new RuntimeException(
						"Virtual relation with an unthis argument:" + this);
			return (singleHypothesis(((VirtualRelation) relation).holds(
					this.component(1), this.component(2))));
		}

		// Check DOMAIN and RANGE on the fly
		if ((this.relationIfKnown() == Relations.DOMAIN || this
				.relationIfKnown() == Relations.RANGE)
				&& this.component(1) != null) {
			Relation ofRelation = Relations.get(this.component(1));
			if (ofRelation == null)
				return (singleHypothesis(false));
			YagoClass classToCheck = this.relationIfKnown() == Relations.DOMAIN ? ofRelation
					.domain() : ofRelation.range();
			if (!isVariable(this.component(2)))
				return (singleHypothesis(classToCheck.name().equals(
						this.component(2))));
			Literal ret = new Literal(false,
					new ArrayList<String>(components()));
			ret.components.set(2, classToCheck.name());
			return (ret.singleHypothesis(true));
		}

		// Check TYPE on the fly for literals
		if (this.relationIfKnown() == Relations.TYPE
				&& this.component(1) != null && this.component(2) != null) {
			YagoClass c = YagoClasses.get(this.component(2));
			if (c == YagoClasses.ENTITY) {
				return (singleHypothesis(true));
			}
			if (c != null && c instanceof LiteralYagoClass) {
				return (singleHypothesis(((LiteralYagoClass) c)
						.hasInstance(this.component(1)) != -1));
			}
			if (c != null
					&& (c == YagoClasses.RELATION || c
							.isSubClassOf(YagoClasses.RELATION))) {
				Relation rel = Relations.get(this.component(1));
				return (singleHypothesis(rel != null && rel.isInstanceOf(c)));
			}
		}

		// Relations for which there is a closed world assumption
		// data does not exist == answer is NO
		if (!Config.isRelationToLearn(component(0))) {
			return (closedWorldIterator(this.instances()));
		}

		// Relations that can be both in the YAGO database and in the hypotheses
		// Any literal will have at least one grounding, either in the YAGO
		// database or in the hypotheses.
		Iterator<Hypothesis> result = PeekIterator.emptyIterator();
		;

		// DisambiguatedAs maps non-polysems to themselves
		if (this.relationIfKnown() == DISAMBIGUATEDAS) {
			// If arg1 is not a polysem, map to yourself
			if (!isVariable(this.component(1)) && !isPolysem(this.component(1))) {
				if (!isVariable(this.component(2)))
					return (singleHypothesis(this.component(1).equals(
							this.component(2))));
				Literal ret = new Literal(false, new ArrayList<String>(
						components()));
				ret.components.set(2, this.component(1));
				return (ret.singleHypothesis(true));
			}
			// If arg2 is there, add the trivial mapping
			if (!isVariable(this.component(2)) && isVariable(this.component(1))) {
				if (trace())
					Announce.message(
							"Adding trivial disambiguation",
							new Hypothesis(true, DISAMBIGUATEDAS, this
									.component(2), this.component(2)));
				result = new PeekIterator.ElementaryPeekIterator<Hypothesis>(
						new Hypothesis(true, DISAMBIGUATEDAS,
								this.component(2), this.component(2)));
			}
		}

		// Query YAGO
		// If our polarity is TRUE, we are only looking for false facts.
		// These cannot be found in YAGO
		if (polarity == false)
			result = new CombinedIterator<Hypothesis>(result, this.instances());
		else if (trace())
			Announce.message("Polarity is positive, don't ask YAGO");

		// Query the hypotheses
		return (new CombinedIterator<Hypothesis>(result, HypothesisStore
				.matchingUnhappyHypotheses(this).iterator()));
	}

	// ----------------------------------------------------------------------------
	// Read a literal from a string
	// ----------------------------------------------------------------------------

	/** Normalizes an identifier */
	protected static String normalize(String a) {
		a = a.trim();
		if (a.startsWith("$"))
			a = "?" + a.substring(1);
		if (a.equals(a.toUpperCase()) && !a.startsWith("?") && a.length() == 1)
			a = "?" + a;
		if (!a.matches("\\??[\\w,]+"))
			Announce.error("Malformed literal component:", a);
		return (a);
	}

	/** Constructs a literal */
	public static Literal parse(String lit) {
		lit = lit.trim();
		if (!lit.matches(".+\\(.*\\)"))
			Announce.error("Malformed literal:", lit);
		boolean polarity = !lit.startsWith("!") && !lit.startsWith("-")
				&& !lit.startsWith("~");
		List<String> parts = new ArrayList<String>(4);
		String rel = normalize(lit
				.substring(polarity ? 0 : 1, lit.indexOf('(')));
		if (!isVariable(rel)) {
			if (Relations.get(rel) != null) {
				rel = Relations.get(rel).name();
			} else {
				if (!Config.isRelationToLearn(rel))
					Announce.error("Unknown relation", rel,
							"must be declared in relationToLearn");
				rel = Config.normalizeRelationToLearn(rel);
			}
		}
		parts.add(rel);
		String[] args = lit.substring(lit.indexOf('(') + 1, lit.length() - 1)
				.replace(",_", "@@").split("\\s*,\\s*");
		for (int i = 0; i < args.length; i++) {
			args[i] = normalize(args[i].replace("@@", ",_"));
		}
		if (Relations.get(rel) == FLAGSET) {
			if (args.length != 1)
				Announce.error("flagSet(X) must have one argument:", lit);
			args[0] = args[0].toUpperCase();
		}
		parts.addAll(Arrays.asList(args));
		return (new Literal(polarity, parts));
	}

	// ----------------------------------------------------------------------------
	// Read hypotheses from the database
	// ----------------------------------------------------------------------------

	/**
	 * Converts a database line <pattern, arg1, arg2, count> to a Hypothesis
	 * object
	 */
	protected static ResultWrapper<Hypothesis> patternOccurrencesWrapper = new ResultWrapper<Hypothesis>() {

		/** Returns a GroundedAtom for a database line */
		public Hypothesis wrap(ResultSet row) throws Exception {
			Relation relation = PATTERNOCCURRENCES;
			String pattern = row.getString(1);
			String arg1 = row.getString(2);
			String arg2 = row.getString(3);
			int value = row.getInt(4);
			return new Hypothesis(value, relation, pattern, arg1, arg2);
		}
	};

	/**
	 * Converts a database line <id, relation, arg1, arg2, value> to a
	 * Hypothesis object
	 */
	public static ResultWrapper<Hypothesis> yagoWrapper = new ResultWrapper<Hypothesis>() {

		/** Returns a GroundedAtom for a database line */
		public Hypothesis wrap(ResultSet row) throws Exception {
			Relation relation = Relations.get(row.getString(2));
			if (relation == null) {
				Announce.error("Unknown relation in database",
						row.getString(1), row.getString(2), row.getString(3),
						row.getString(4));
			}
			String arg1 = row.getString(3);
			String arg2 = row.getString(4);
			//double value = row.getDouble(5); // YAGO2 now longer has confidences
			double value = 0.5;
			return new Hypothesis(value, relation, arg1, arg2);
		}
	};

	/**
	 * Converts a hypothesis line <wordInContext, entity, confidence> to a Fact
	 * object
	 */
	public static ResultWrapper<Hypothesis> disambiguationPriorWrapper = new ResultWrapper<Hypothesis>() {

		/** Returns a GroundedAtom for a database line */
		public Hypothesis wrap(ResultSet row) throws Exception {
			Relation relation = DISAMBIGUATIONPRIOR;
			String arg1 = row.getString(1);
			String arg2 = row.getString(2);
			double value = row.getDouble(3);
			return new Hypothesis(value, relation, arg1, arg2);
		}

	};

	/** TRUE if this literal has instances */
	protected Truth hasInstances = Truth.UNKNOWN;

	/** Loads all instantiations of this literal from the database */
	public PeekIterator<Hypothesis> instances() throws SQLException {
    if(hasInstances==Truth.FALSE) return(PeekIterator.emptyIterator());
    String table = "facts";
    List<String> columnNames = Arrays.asList("relation", "arg1", "arg2");
    List<String> componentNames = components();
    ResultIterator.ResultWrapper<Hypothesis> wrapper = yagoWrapper;
    if (relationIfKnown() == DISAMBIGUATIONPRIOR) {
      table = "disambiguationPrior";
      columnNames = Arrays.asList("WordInContext", "Entity");
      componentNames = components.subList(1, components.size());
      wrapper = disambiguationPriorWrapper;
    } else if (relationIfKnown() == PATTERNOCCURRENCES) {
      table = "patternOccurrences";
      columnNames = Arrays.asList("patternID", "arg1", "arg2");
      componentNames = components.subList(1, components.size());
      wrapper = patternOccurrencesWrapper;
    }
    ;
    StringBuilder sql = new StringBuilder("SELECT * FROM ").append(table).append(" WHERE ");
    for (int i = 0; i < columnNames.size(); i++) {
      if (isVariable(componentNames.get(i))) continue;
      sql.append(columnNames.get(i)).append('=');
      sql.append(Config.database.format(componentNames.get(i)));
      sql.append(" AND ");
    }
    if (relationIfKnown() == PATTERNOCCURRENCES && Config.delimitPOC != null) sql.append(Config.delimitPOC);
    if (sql.substring(sql.length() - 5).equals(" AND ")) sql.setLength(sql.length() - 5);
    if (sql.substring(sql.length() - 7).equals(" WHERE ")) sql.setLength(sql.length() - 7);
    PeekIterator<Hypothesis> returnMe=Config.database.query(sql, wrapper);
    hasInstances=Truth.of(returnMe.hasNext());
    return (returnMe);
  }

	/** TRUE for a literal that has instances */
	public boolean hasInstances() throws SQLException {
		if (hasInstances != Truth.UNKNOWN)
			return (hasInstances.asBoolean());
		PeekIterator<Hypothesis> it = instances();
		boolean retVal = it.hasNext();
		it.close();
		return (retVal);
	}

	/** Test method */
	public static void main(String[] args) throws Exception {
		D.p(parse(" !abc(?de,EF,gh, X) "));
	}
}
