package sofie.maxsat;
import java.util.Collection;
import java.util.Set;
import java.util.TreeSet;

/**
 * Class HypothesisWithMeasures
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *
 * A Hypothesis that registers how many ground rules are unsatisfied.
 *  
 * @author Fabian M. Suchanek
 */
public class Measures {
  /** Index in arrays*/
  protected static final int TRUE=1;
  /** Index in arrays*/
  protected static final int FALSE=0;
  /** Non-unit rules*/
  protected double[] nonUnitWeight=new double[2];
  /** Unit rules*/
  protected double[] unitWeight=new double[2];
  /** Johnsons Measures (for debugging)*/
  protected double[] johnson=new double[2];
  /** Other related facts */
  protected Set<Hypothesis> relatedHypotheses=new TreeSet<Hypothesis>();
    
  @Override  
  public String toString() {     
    StringBuilder b=new StringBuilder();
    b.append(" Unit false:").append(unitWeight[FALSE]);
    b.append(" Unit true:").append(unitWeight[TRUE]);
    b.append(" Non-unit false:").append(nonUnitWeight[FALSE]);
    b.append(" Non-unit true:").append(nonUnitWeight[TRUE]);
    b.append(" Johnson false:").append(johnson[FALSE]);
    b.append(" Johnson true:").append(johnson[TRUE]);
    return(b.toString());
  } 
  
  /** TRUE if the truth value of the hypothesis is determined by DUC propagation*/
  public boolean isDetermined() {
    if(unitWeight[TRUE]>=nonUnitWeight[FALSE]+unitWeight[FALSE]) return(true);
    if(unitWeight[FALSE]>=nonUnitWeight[TRUE]+unitWeight[TRUE]) return(true);
    return(false);
  }
  
  /** Double value that estimates how urgent this hypothesis should be set*/
  public double priority() {
    if(isDetermined()) return(Double.MAX_VALUE);
    return(Math.abs(unitWeight[TRUE]-unitWeight[FALSE]));
  }

  /** Heuristic truth value*/
  public boolean heuristicTruthValue() {
    return(unitWeight[TRUE]>unitWeight[FALSE]);      
  }

  /** Deterministic truth value (if isDetermined)*/
  public boolean determinedTruthValue() {
    return(unitWeight[TRUE]>=nonUnitWeight[FALSE]+unitWeight[FALSE]);
  }

  /** What Johnson would say (for debugging purposes)*/
  public int johnsonsTruthValue() {
    return(johnson[TRUE]>johnson[FALSE]?TRUE:FALSE);      
  }

  /** Sets the truth value*/
  public boolean truthValue() {
    if(isDetermined()) return(determinedTruthValue());
    return(heuristicTruthValue());
  }
  
  /** Adds the effect of a grounded rule*/
  public void addMeasures(int numFreeLiterals,int numUnsatisfied, boolean literalPolarity, double weight) {
    int polarity=literalPolarity?TRUE:FALSE;
    if(numFreeLiterals-numUnsatisfied==1) unitWeight[polarity]+=weight;
    else nonUnitWeight[polarity]+=weight;
    johnson[polarity]+=weight/(1<<(numFreeLiterals-numUnsatisfied));
  }
  
  /** Registers a free hypothesis that appears together with this hypothesis in a rule*/
  public void addRelatedHypothesis(Hypothesis f) {
    if(!f.equals(this)) relatedHypotheses.add(f);      
  }   

  /** Registers free hypotheses that appear together with this hypothesis in a rule*/
  public void addRelatedHypotheses(Collection<Hypothesis> fs) {
    for(Hypothesis hyp : fs) addRelatedHypothesis(hyp);
  }   
  /** returns the related hypotheses*/
  public Set<Hypothesis> relatedHypotheses() {
    return(relatedHypotheses);
  }
}
