package sofie.parsing;

import java.io.File;
import java.lang.reflect.Field;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.database.Database;
import javatools.database.ResultIterator;
import javatools.datatypes.ArrayQueue;
import javatools.datatypes.FinalSet;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.FileSet;
import javatools.parsers.Char;
import javatools.parsers.Name;
import javatools.parsers.NumberParser;
import javatools.parsers.PlingStemmer;
import basics.datagraph.DataGraphManager;
import basics.datagraph.YagoClasses;
import basics.datagraph.Relations;

/**
 * Class Parser
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *      
 * This class parses the natural language documents that serve as input for SOFIE.
 * 
 * @author Fabian M. Suchanek
 */
public class Parser {

  // ---------------------------------------------------------------------
  //           Configuration
  // ---------------------------------------------------------------------

  // These parameters are loaded from the ini-file.
  // The name of the Java variable must correspond to the parameter name in the 
  // ini-file. It must be in upper case in the Java source.
  // If a parameter is not found in the ini-file, the default value is used.
  // The default value is specified in the following for each parameter.

  /** Number of possible meanings of a word that are taken into account*/
  private int NUMMEANINGS = 3;

  /** Take into account words that have multiple meanings*/
  private boolean POLYSEMS = true;

  /** Keep old patterns, do not overwrite them*/
  private boolean KEEPOLDPATTERNS = false;

  /** Maximal length of a pattern in characters*/
  private int MAXPATTERNCHARS = 100;

  /** Minimal length of a pattern in tokens*/
  private int MINPATTERNTOKENS = 1;

  /** Maximal length of a pattern in tokens*/
  private int MAXPATTERNTOKENS = 8;

  /** DEBUGGING: If you would like to know why a specific word is disambiguated 
   * in a particular way, add the word here*/
  private Collection<String> traceWords = new FinalSet<String>(/** Add here*/
  );

  /** Points to the database*/
  private Database database;

  // ---------------------------------------------------------------------
  //                  Global data
  // ---------------------------------------------------------------------

  /** Maintains the Id of newly added patterns*/
  private int currentPatternId = 0;

  /** Maintains the Id of added patternOccurrences*/
  private int currentPatternOccurrenceId = 0;

  /** Holds the Id of the current document*/
  private int currentDocumentId = 0;

  // ---------------------------------------------------------------------
  //                  Helper Methods
  // ---------------------------------------------------------------------

  /** Returns the string "table.relation='means' AND table.arg1=... AND table.arg2=..." */
  protected String means(String table, String arg1, String arg2) {
    if (table == null) table = "";
    else table = table + ".";
    Database db = database;
    return ("(" + table + "relation=" + db.format(Relations.MEANS) + " OR " + table + "relation=" + db.format(Relations.FAMILYNAME) + ") AND "
        + table + (arg1 != null ? "arg1=" + db.format(arg1) : "arg2=" + db.format(arg2)));
  }

  /** Executes a database update. Hook here for debugging*/
  protected int update(String sql) throws SQLException {
    Announce.debugMsg(sql);
    return (database.executeUpdate(sql));
  }

  /** Tells whether a word is a pronoun*/
  public boolean isPronoun(String s) {
    s = s.toLowerCase();
    return (s.equals("he") || s.equals("she") || s.equals("his") || s.equals("her"));
  }

  /** Returns a normal word for a Yago identifier*/
  public String nameOfEntity(String y) {
    if (y.startsWith("wikicategory_")) return (y.substring(13).replace('_', ' '));
    if (y.startsWith("wordnet_")) {
      return (y.substring(8, y.lastIndexOf('_')).replace('_', ' '));
    }
    return (y.replace('_', ' '));
  }

  // ---------------------------------------------------------------------
  //                  Type checks
  // ---------------------------------------------------------------------

  /** Contains all words from WordNet*/
  protected static Set<String> commonWords = new TreeSet<String>();

  /** Loads the words of WordNet*/
  public static void loadWordNet(File folder) {
    if (commonWords.size() != 0) return;
    try {
      for (String line : new FileLines(new File(folder, "wn_s.pl"), "Loading WordNet")) {
        //s(100001740,1,'entity',n,1,11).
        String word = line.split("'*,'*")[2];
        if (Name.isName(word) || word.contains(" ")) continue;
        commonWords.add(word);
      }
    } catch (Exception e) {
      Announce.failed();
      return;
    }
  }

  /** Returns words*/
  public static boolean isCommonword(String w) {
    return (commonWords.contains(w.toLowerCase()) || commonWords.contains(PlingStemmer.stem(w.toLowerCase())));
  }

  /** Constant for interestingTypes()*/
  public static final int DOMAIN = 0;

  /** Constant for interestingTypes()*/
  public static final int RANGE = 1;

  /** Checks whether the pair of entities is interesting. */
  public static boolean interestingTypes(Token entity, int domainRange) {
    if (entity instanceof Token.Literal) {
      if (domainRange == DOMAIN) return (false);
      return (true);
    }
    if (entity instanceof Token.CommonIndividual) {
      switch (((Token.CommonIndividual) entity).nameType()) {
        case POLYSEM:
          return (true);
        case UNKNOWN:
          return (false);
        case SEMICANONIC:
          return (false);
        case CANONIC:
          return (true);
      }
    }
    return (true);
  }

  // ---------------------------------------------------------------------
  //                  Disambiguation
  // ---------------------------------------------------------------------  

  /** Stores all words of the current document with their frequency*/
  protected Map<String, Integer> wordsOfCurrentDocument;

  /** Checks whether an entity is a person and redirects the familyname */
  protected void replaceFamilyName(Token token, List<Token> tokens) throws SQLException {
    boolean trace = traceWords.contains(token.token()) || token.token().split(" |_").length > 1 && traceWords.contains(token.token().split(" |_")[1]);
    if (trace) Announce.message("Checking as potential person name:", token);
    // Check whether the token could be a person name
    if (!token.isMultiWord()) return;
    if (!(token instanceof Token.CommonIndividual)) return;
    if (((Token.CommonIndividual) token).nameType() != Token.NameType.UNKNOWN) return;
    if (!Name.couldBePersonName(token.token())) return;
    if (trace) Announce.message("Token could be a person name");
    // Check whether the family name appears
    String family = new Name.PersonName(token.token()).familyName();
    if (family == null) return;
    int familyNamePos = tokens.indexOf(new Token(family));
    if (familyNamePos == -1 || !(tokens.get(familyNamePos) instanceof Token.CommonIndividual)) return;
    if (trace) Announce.message("Familyname of the token appears as", tokens.get(familyNamePos));
    // Check whether one meaning is a person
    for (String meaning : disambiguate(token)) {
      StringBuilder sql = new StringBuilder("SELECT * FROM facts WHERE ");
      sql.append("arg1=").append(database.format(meaning)).append(" AND ");
      sql.append("relation=").append(database.format(Relations.TYPE)).append(" AND ");
      sql.append("arg2=").append(database.format(YagoClasses.PERSON));
      // If so, redirect the familyname
      if (database.exists(sql)) {
        if (trace) Announce.message("Token is a person. I am redirecting the familyname to", token);
        ((Token.CommonIndividual) tokens.get(familyNamePos)).redirectTo((Token.CommonIndividual) token);
        // Since all tokens are unique, it suffices to redirect one family name token
        break;
      }
    }
  }

  /** Returns the context of an entity from the database*/
  public Set<String> ontologicalContext(String entity) throws SQLException {
    TreeSet<String> result = new TreeSet<String>();
    result.addAll(database.query("SELECT arg2 FROM facts WHERE arg1=" + database.format(entity), ResultIterator.StringWrapper).asList());
    result.addAll(database.query("SELECT arg1 FROM facts WHERE arg2=" + database.format(entity), ResultIterator.StringWrapper).asList());
    result.add(entity);
    return (result);
  }

  /** Generates the disambiguation prior and stores it in the database*/
  protected void generateDisambiguationPrior(Token token, Set<String> meanings) throws SQLException {
    String word = token.original();
    int[] topOverlaps = new int[NUMMEANINGS];
    String[] topEntities = new String[NUMMEANINGS];
    Arrays.fill(topOverlaps, -1);

    // Consider all possible meanings of the word
    for (String entity : meanings) {

      // Compute the overlap of the ontological context and the word
      int overlap = 0;
      if (D.indexOf(entity, (Object[]) topEntities) != -1) continue;
      for (String contextEntity : ontologicalContext(entity)) {
        for (String part : nameOfEntity(contextEntity).split(" ")) {
          while (part.length() > 0 && !Character.isLetterOrDigit(part.charAt(0)))
            part = part.substring(1);
          while (part.length() > 0 && !Character.isLetterOrDigit(Char.last(part)))
            part = Char.cutLast(part);
          if (part.equalsIgnoreCase(word) || word.indexOf(part) != -1 || part.length() < 3 || Name.isStopWord(part.toLowerCase())) continue;
          overlap += D.getOrZero(wordsOfCurrentDocument, part);
        }
      }
      overlap += D.getOrZero(wordsOfCurrentDocument, entity);

      // Add the current pair of entity and counter to our list   
      // Overwrite some other entity that has a lower counter
      for (int i = 0; i < NUMMEANINGS; i++) {
        if (topOverlaps[i] < overlap) {
          topOverlaps[i] = overlap;
          topEntities[i] = entity;
          break;
        }
      }
    }

    // Normalize the counters
    double sum = 0;
    for (int i = 0; i < NUMMEANINGS && topEntities[i] != null; i++) {
      sum += topOverlaps[i];
    }
    if (sum == 0) {
      sum = 100;
      Arrays.fill(topOverlaps, 1);
    }

    // Write the counters
    for (int i = 0; i < NUMMEANINGS && topEntities[i] != null; i++) {
      double value = topOverlaps[i] / sum;
      update("INSERT INTO disambiguationPrior VALUES(" + database.format(token.token()) + ", " + database.format(topEntities[i]) + ", " + value + ")");
    }
  }
  
  /** Determines whether the token is ambigous, unambiguos or void. 
   * Returns set of possible meanings. */
  public Collection<String> disambiguate(Token token1) throws SQLException {
    if (traceWords.contains(token1.token())) Announce.message("Disambiguating token", token1);
    // Check whether there is something to do
    if (!(token1 instanceof Token.CommonIndividual)) return (Arrays.asList(token1.token()));
    Token.CommonIndividual token = ((Token.CommonIndividual) token1);
    switch (token.nameType()) {
      case CANONIC:
        return (Arrays.asList(token1.token()));
      case POLYSEM:
        return (Arrays.asList(token1.token()));
      case VOID:
        return (Arrays.asList());
      case UNKNOWN: // handle in this method
      case SEMICANONIC: // handle in this method
    }
    String original = token.token();

    // Get meanings together with their confidence scores
    String name = '\"' + original + '\"';
    List<String[]> possibleMeaningsWithConfidence = database.query(
        "SELECT arg2 FROM facts WHERE relation=" + database.format(Relations.MEANS) + " AND arg1=" + database.format(name),
        ResultIterator.StringsWrapper).asList();

    // If it's a company name, try asking YAGO without the "inc."
    if (possibleMeaningsWithConfidence.size() == 0 && token instanceof Token.Company) {
      Name.CompanyName company = new Name.CompanyName(original);
      name = '"' + company.name() + '"';
      if (traceWords.contains(token1.token())) Announce.message("No meaning found, trying", name);
      possibleMeaningsWithConfidence = database.query(
          "SELECT arg2 FROM facts WHERE relation=" + database.format(Relations.MEANS) + " AND arg1=" + database.format(name),
          ResultIterator.StringsWrapper).asList();
      if (possibleMeaningsWithConfidence.size() == 0) {
        possibleMeaningsWithConfidence = database.query(
            "SELECT arg2 FROM facts WHERE relation=" + database.format(Relations.MEANS) + " AND arg1="
                + database.format(Char.capitalize(name)), ResultIterator.StringsWrapper).asList();
      }
    }

    // Collect meanings, kill wordnet meanings and single word person meanings
    SortedSet<String> possibleMeanings = new TreeSet<String>();
    String preferredMeaning = null;
    for (String[] meanConf : possibleMeaningsWithConfidence) {
      // Sort out wordnet meanings
      if (meanConf[0].startsWith("wordnet")) continue;
      // Add the meaning
      possibleMeanings.add(meanConf[0]);
      // If the confidence value is 0.95..., this is the preferred meaning in YAGO
      // (this is a hack)
      //if (meanConf[1].startsWith("0.95")) preferredMeaning = meanConf[0];
    }
    if (traceWords.contains(token1.token())) Announce.message("Preferred Meaning:", preferredMeaning, "Other Meanings:", possibleMeanings);

    // If there are no meanings, give up
    if (possibleMeanings.size() == 0) {
      token.makeVoid();
      return (Arrays.asList());
    }

    // If the word is the exact name of an entity, we're done
    String entityName = token.token().replace(' ', '_');
    if (possibleMeanings.contains(entityName)) {
      if (traceWords.contains(token1.token())) Announce.message("Token is name of article");
      token.disambiguateAs(entityName);
      return (Arrays.asList(token.token()));
    }

    // If the word was semi-canonic, it should have a unique meaning
    if (token.nameType() == Token.NameType.SEMICANONIC) {
      if (possibleMeanings.size() == 1) {
        if (traceWords.contains(token1.token())) Announce.message("Token is an immediate redirect to an article");
        token.disambiguateAs(possibleMeanings.first());
        return (Arrays.asList(token.token()));
      }
      if (traceWords.contains(token1.token())) Announce.message("Token is an immediate redirect to an article but goes nowhere");
      token.makeVoid();
      return (Arrays.asList());
    }

    // If there is a meaning in Wordnet, it's probably that one and we don't want it
    if (isCommonword(original)) {
      if (traceWords.contains(token1.token())) Announce.message("Token is a common word");
      token.makeVoid();
      return (Arrays.asList());
    }

    // If there is a unique meaning, we're done
    if (possibleMeanings.size() == 1) {
      if (traceWords.contains(token1.token())) Announce.message("Token has only one meaning");
      token.disambiguateAs(possibleMeanings.first());
      return (Arrays.asList(token.token()));
    }


    // For true polysems, come here
    if (!POLYSEMS) {
      if (traceWords.contains(token1.token())) Announce.message("Token is polysem, but polysems are switched off");
      token.makeVoid();
      return (Arrays.asList());
    }

    // OK, it has multiple meanings
    token.makePolysem(currentDocumentId);
    if (traceWords.contains(token1.token())) Announce.message("Token is polysem");
    generateDisambiguationPrior(token, possibleMeanings);
    return (possibleMeanings);
  }

  // ---------------------------------------------------------------------
  //                  Saving Patterns
  // ---------------------------------------------------------------------

  /** Returns a String representation of a pattern (excluding start and end)*/
  public String simplifiedPattern(List<Token> tokens, int start, int end) {
    if (start < 0) return (null);
    StringBuilder result = new StringBuilder();
    boolean wasPlaceholder = false;
    boolean hasGoodToken = false;

    // Collect all tokens, replace replacable tokens by '@'
    for (int i = start + 1; i < end && i < tokens.size(); i++) {
      Token token = tokens.get(i);
      if (token.replaceMe()) {
        if (!wasPlaceholder) result.append('@');
        wasPlaceholder = true;
      } else {
        result.append(token.token()).append(' ');
        if (token instanceof Token.Word) hasGoodToken = true;
        wasPlaceholder = false;
      }
    }

    // If the pattern did not have good tokens, ignore it
    if (!hasGoodToken) return (null);
    //  Avoid inserting patterns into the database that are too long
    if (result.length() > MAXPATTERNCHARS) return (null);
    return (result.toString().trim());
  }

  /** Stores a pattern in the database */
  public void savePattern(String simplifiedPattern, String arg1, String arg2) throws SQLException {
    // Find the pattern
    Integer patternId = database.queryValue("SELECT patternId FROM patterns WHERE pattern=" + database.format(simplifiedPattern),
        ResultIterator.IntegerWrapper);

    // If the pattern does not exist, add it
    if (patternId == null) {
      patternId = ++currentPatternId;
      update("INSERT INTO patterns VALUES(" + patternId + ", " + database.format(simplifiedPattern) + ")");
      currentPatternOccurrenceId++;
      update("INSERT INTO patternOccurrences VALUES(" + patternId + ", " + database.format(arg1) + ", " + database.format(arg2) + ", 1, "
          + currentPatternOccurrenceId + ")");
    } else {
      // If the pattern does exist, try to update this very instance
      int updateCount = update("UPDATE patternOccurrences SET count=count+1 WHERE patternId=" + patternId + " AND arg1=" + database.format(arg1)
          + " AND arg2=" + database.format(arg2));
      // If the pattern exists, but not this very instance, create a new instance  
      if (updateCount == 0) {
        currentPatternOccurrenceId++;
        update("INSERT INTO patternOccurrences VALUES(" + patternId + ", " + database.format(arg1) + ", " + database.format(arg2) + ", 1, "
            + currentPatternOccurrenceId + ")");
      }
    }
  }

  /** Checks and stores a binary pattern to the database*/
  public void saveBinaryPattern(List<Token> tokens, int start, int end) throws SQLException {
    // Check whether the pattern is worth being saved
    if (end - start <= MINPATTERNTOKENS) return;
    Token arg1 = tokens.get(start);
    Token arg2 = tokens.get(end);
    if (arg1.equals(arg2)) return;
    if (arg1 instanceof Token.Literal) return;

    // Get the pattern
    String simplifiedPattern = simplifiedPattern(tokens, start, end);

    // If the pattern is a Wikipedia category, include also the following two words.
    // This serves for patterns like "Albert_Einstein CATEGORY 1879 births"
    if (tokens.get(start + 1) == Token.CATEGORY) {
      simplifiedPattern = simplifiedPattern(tokens, start, end + 2);
    }
    if (simplifiedPattern == null) return;

    // Disambiguate the entities
    if (disambiguate(arg1).isEmpty()) return;
    if (disambiguate(arg2).isEmpty()) return;
    if (arg1.token().equals(arg2.token())) return;

    // Do a last type check
    if (!interestingTypes(arg1, 0)) return;
    if (!interestingTypes(arg2, 1)) return;

    // ... and save the pattern
    savePattern(simplifiedPattern, arg1.token(), arg2.token());
  }

  // ---------------------------------------------------------------------
  //                  Finding Patterns
  // ---------------------------------------------------------------------

  /** Collects all words in the current document*/
  public static Map<String, Integer> collectWordsFromTokens(List<Token> tokens) {
    Map<String, Integer> words = new TreeMap<String, Integer>();
    for (Token t : tokens) {
      for (String s : t.token().split(" "))
        D.addKeyValue(words, s, 1);
    }
    return (words);
  }

  /** Finds and stores the patterns from a list of tokens*/
  public void findPatterns(List<Token> tokens) throws SQLException {
    wordsOfCurrentDocument = collectWordsFromTokens(tokens);
    Queue<Integer> entityPositions = new ArrayQueue<Integer>();
    Announce.progressStart("Finding patterns", tokens.size());
    for (int i = 0; i < tokens.size(); i++) {
      Announce.progressAt(i);
      Token token = tokens.get(i);
      replaceFamilyName(token, tokens);
      if (token instanceof Token.Separator) {
        entityPositions.clear();
        continue;
      }
      if (token.isInteresting()) {
        entityPositions.add(i);
        while (i - entityPositions.peek() > MAXPATTERNTOKENS)
          entityPositions.poll();
        for (Integer j : entityPositions) {
          saveBinaryPattern(tokens, j, i);
        }
      }
    }
    wordsOfCurrentDocument = null;
    Announce.progressDone();
  }

  // ---------------------------------------------------------------------
  //                  Main
  // ---------------------------------------------------------------------

  /** Resets all patterns*/
  public void resetPatterns() throws SQLException {
    if (KEEPOLDPATTERNS) {
      try {
        currentPatternId = database.queryValue("SELECT MAX(patternId) FROM patterns", ResultIterator.IntegerWrapper) + 1;
      } catch (NullPointerException e) {
      }
      try {
        currentPatternOccurrenceId = database.queryValue("SELECT MAX(id) FROM patternOccurrences", ResultIterator.IntegerWrapper) + 1;
      } catch (NullPointerException e) {
      }
      Announce.message("Existing patterns have been kept");
      return;
    }
    Announce.doing("Resetting patterns");
    database.createTable("patternOccurrences", "patternid", java.sql.Types.INTEGER, "arg1", java.sql.Types.VARCHAR, "arg2", java.sql.Types.VARCHAR,
        "count", java.sql.Types.FLOAT, "id", java.sql.Types.INTEGER);
    database.createIndices("patternOccurrences", "patternID", "arg1", "arg2");
    database.createTable("patterns", "patternid", java.sql.Types.INTEGER, "pattern", java.sql.Types.VARCHAR);
    database.createIndex("patterns", true, "pattern");
    database.createTable("disambiguationPrior", "wordInContext", java.sql.Types.VARCHAR, "entity", java.sql.Types.VARCHAR, "value",
        java.sql.Types.FLOAT);
    database.createIndex("disambiguationPrior", false, "wordInContext");
    database.createIndex("disambiguationPrior", false, "entity");
    Announce.done();
  }

  /** Parses the files */
  public Parser(File iniFile, boolean keepPatterns) throws Exception {
    // Initialize all parameters from the ini-file
    Parameters.init(iniFile);
    DataGraphManager.init();

    for (Field field : Parser.class.getDeclaredFields()) {
      String parameterName = field.getName();
      if (parameterName.equals(parameterName.toUpperCase()) && Parameters.isDefined(parameterName)) {
        if (field.getType() == Integer.class || field.getType() == int.class) field.set(this, Parameters.getInt(parameterName));
        else if (field.getType() == Boolean.class || field.getType() == boolean.class) field.set(this, Parameters.getBoolean(parameterName));
        else if (D.indexOf(List.class, (Object[]) field.getType().getInterfaces()) != -1) field.set(this, Parameters.getList(parameterName));
        else field.set(this, Parameters.get(parameterName));
      }
    }
    KEEPOLDPATTERNS=keepPatterns;
    
    // Connect to database
    Announce.doing("Connecting to database");
    database = Parameters.getDatabase();
    Announce.done();
    Announce.message("Connected to", database);
    resetPatterns();

    // Load WordNet if present
    if (Parameters.isDefined("wordnet")) {
      loadWordNet(Parameters.getFile("wordnet"));
    } else {
      Announce.warning("To improve performance, download the Prolog version of WordNet\n" + "and write 'wordnet=<wordnetFolder>' in the ini-file.");
    }

  }

  public void parse(List<String> filenames, int start) throws Exception {
    // Parse all files
    currentDocumentId=0;
    Announce.doing("Finding file to start with");
    for (String filename : filenames) {
      for (File f : new FileSet(filename)) {
        if(++currentDocumentId==start) Announce.done();
        else if (currentDocumentId<start) continue;
        Announce.doing("Parsing", f, currentDocumentId);
        findPatterns(Tokenizer.autodetectTokenize(f));
        Announce.done();
      }
    }
  }
  
  public void close() {
    Announce.doing("Closing database");
    database.close();
    Announce.done();
  }

  /** Finds patterns in a file*/
  public static void main(String[] args) throws Exception {

    // Prepare arguments
    if (args.length == 0 || Announce.isHelp(args[0])) {
      Announce.help("     sofie.parsing.Parser [<ini-file>] [<start>] <files>...\n", "Parses the files and stores the patterns in the database.",
          "<ini-file>    an initialization file that defines the database and other parameters.", "              Defaults to sofie.ini .",
          "<start>       the number of the file to start with if interrupted (defaults to 1)",
          "<files>       a list of files (possibly given with wildcards) that will be parsed.");
    }
    List<String> filenames = new ArrayList<String>(Arrays.asList(args));
    File iniFile = new File("sofie.ini");
    if (args[0].endsWith(".ini")) {
      iniFile = new File(filenames.get(0));
      filenames.remove(0);
    }
    int start = 1;
    if (NumberParser.isInt(filenames.get(0))) {
      start = NumberParser.getInt(filenames.get(0));
      filenames.remove(0);
    }
    Parser parser=new Parser(iniFile, start > 1);
    parser.parse(filenames, start);
    parser.close();
  }

}
