package sofie.parsing;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.filehandlers.FileSet;
import javatools.filehandlers.UTF8Reader;
import javatools.parsers.Char;
import javatools.parsers.DateParser;
import javatools.parsers.Name;
import javatools.parsers.NumberParser;

/**
 * Class Tokenizer
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *
 * Tokenizes a String.
 *  
 * @author Fabian M. Suchanek
 */

public class Tokenizer {
  
  // ------------------------------------------------------------------------------------------
  //                       Helper Methods
  // ------------------------------------------------------------------------------------------

  /** Returns a token for a predefined word*/
  public static Token canonicalizeKnownWord(Token s) { 
    if(s.token().equals("an")) return(new Token.StopWord("a"));
    if(Name.isStopWord(s.token().toLowerCase())) return(new Token.StopWord(s.token().toLowerCase()));
    if(Name.isLanguage(s.token())) return(new Token.CanonicProperName(s.token()+"_language"));
    if(Name.isNationality(s.token())) return(new Token.SemiCanonicProperName(Name.nationForNationality(s.token())));
    if(Name.isUSState(s.token())) return(new Token.USState(s.token()));
    if(Name.unabbreviateUSState(s.token())!=null) return(new Token.USState(Name.unabbreviateUSState(s.token())));
    return(s);
  }
  
  // ------------------------------------------------------------------------------------------
  //                       Pattern Runner
  // ------------------------------------------------------------------------------------------

  /** A pattern runner is an object that finds one specific pattern in a string.
  * It knows to what token this pattern shall be mapped.*/
  protected static class PatternRunner {

    /** Token by which an occurence shall be replaced (or NULL if token shall be newly created)*/
    protected Token token;

    /** Type of token*/
    protected Class<? extends Token> tokenType;

    /** Pattern to be found*/
    protected Pattern pattern;

    /** Matcher if running on a specific charsequence (NULL if pattern could not be found anymore)*/
    protected Matcher matcher;

    public PatternRunner(Pattern p, Class<? extends Token> c, Token t) {
      tokenType = c;
      token = t;
      pattern = p;
    }

    public PatternRunner(String p, Class<? extends Token> c, Token t) {
      this(Pattern.compile(p), c, t);
    }

    public PatternRunner(String p, int flags, Class<? extends Token> c, Token t) {
      this(Pattern.compile(p, flags), c, t);
    }

    public PatternRunner(Pattern p, Token t) {
      this(p, t.getClass(), t);
    }

    public PatternRunner(String p, int flags, Token t) {
      this(Pattern.compile(p, flags), t.getClass(), t);
    }

    public PatternRunner(String p, Token t) {
      this(Pattern.compile(p), t);
    }

    public PatternRunner(String p, int flags, Class<? extends Token> c) {
      this(p, flags, c, null);
    }

    public PatternRunner(String p, Class<? extends Token> c) {
      this(p, c, null);
    }

    public PatternRunner(Pattern p, Class<? extends Token> c) {
      this(p, c, null);
    }

    /** Sets the pattern runner on this char sequence. Returns this*/
    public PatternRunner runOn(CharSequence s) {
      matcher = pattern.matcher(s);
      if (!matcher.find()) matcher = null;
      return(this);
    }

    /** Tells whether this pattern runner can find a better pattern 
     * for this position than the previousBest. Returns 0=no, 1=yes, -1=don't know*/
    public int beatsForPosition(PatternRunner previousBest, int position) {
      if (matcher == null) return (0);
      if (matcher.start() < position) {
        if (!matcher.find(position)) {
          matcher = null;
          return (0);
        }
      }
      if (previousBest == null) return (1);
      if (matcher.start() <= previousBest.matcher.start()) return (1);
      return (-1);
    }

    /** Adds this runner's token to the list of tokens. Changes the current title token if necessary*/ 
    public Token addReplacement(List<Token> tokens, Token currentTitle, Map<Token,Token> uniqueTokens) {
      // Ignore tokens go away
      if (token == Token.IGNORE) return(null);
      // Title tokens change the current title token
      if (tokenType == Token.Title.class) {        
        if(matcher.group(1)!=null) {
          Token result=new Token.CanonicProperName(matcher.group(1).replace(' ', '_'));
          tokens.add(result);
          return(result);
        }
      }
      // Repeat tokens add a separator and the title token
      if (tokenType == Token.Repeat.class) {
        tokens.add(Token.SEPARATOR);
        tokens.add(currentTitle); 
      }
      // If the token is predefined, add that one
      if (token != null) {
        tokens.add(token);
        return(token);
      }
      // Else construct a new token
      String tokenString=matcher.groupCount() == 1 ? matcher.group(1) : matcher.group();      
      if(tokenString==null) return(null);            
      Token result=null;
      try {
        result=tokenType.getConstructor(String.class).newInstance(tokenString);
      } catch (Exception e) { e.printStackTrace(); return(null);}
      result=canonicalizeKnownWord(result);
      // Contract patterns of the form "<city>, <USState>"
      if(result instanceof Token.USState 
          && tokens.size()>2 
          && "/,(".indexOf(tokens.get(tokens.size()-1).token().charAt(0))!=-1
          && tokens.get(tokens.size()-2) instanceof Token.CommonIndividual
          && !tokens.get(tokens.size()-2).token().contains(",")
          && !tokens.get(tokens.size()-2).token().contains("/")
          && !tokens.get(tokens.size()-2).token().contains("(")) {
        result=new Token.SemiCanonicProperName(tokens.get(tokens.size()-2).token()+", "+result.token());
        tokens.remove(tokens.size()-1);
        tokens.remove(tokens.size()-1);
      }
      // Maintain our record of unique tokens
      Token uniqueToken=uniqueTokens.get(result);
      if(uniqueToken==null) uniqueTokens.put(uniqueToken=result,result);
      tokens.add(uniqueToken);
      return(uniqueToken);      
    }

    /** Returns the position after the match*/
    public int newPosition() {
      return (matcher.end());
    }
    
    /** returns the pattern */
    public Pattern pattern() {
      return(pattern);
    }
    
    @Override
    public String toString() {     
      return "Pattern: "+pattern+" Matcher: "+matcher;
    }
  }

  // ------------------------------------------------------------------------------------------
  //                       Default Tokenizer
  // ------------------------------------------------------------------------------------------
  
  /** Standard Pattern runners*/
  protected static final List<PatternRunner> standardPatternRunners = Arrays.asList(
      new PatternRunner("[\\p{L}_]++", Token.Word.class),
    new PatternRunner("\\p{Punct}", Token.Punctuation.class),
    new PatternRunner("[\\.;]", Token.SEPARATOR),    
    new PatternRunner(NumberParser.NUMBERPATTERN, Token.Number.class),
    new PatternRunner("\\b(1[4-9]\\d\\d)\\b", Token.Date.class),
    new PatternRunner(DateParser.DATEPATTERN, Token.Date.class),
    new PatternRunner(Name.safeNamesPattern, Token.ProperName.class),
    new PatternRunner(Name.safePersonNamePattern, Token.Person.class),
    new PatternRunner(Name.safeAbbreviationPattern, Token.ProperName.class),
    new PatternRunner(Name.safeCompanyPattern, Token.Company.class));

  /** Pattern runners that are to be run for this specific type of tokenizer.
   * Default: standardPatternRunners*/
  protected List<PatternRunner> patternRunners() {
    return(standardPatternRunners);
  }
  
  /** Prepares a string for tokenizing.
   * Default: Normalize numbers and dates*/
  protected String prepare(CharSequence s) {
    String result = NumberParser.normalize(DateParser.normalize(s).replaceAll("\\s++", " "));
    return (result);
  }

  /** Main method: Tokenizes a char sequence*/
  public List<Token> tokenize(CharSequence s) {  
    Announce.doing("Mounting Tokenizer");
    Map<Token,Token> uniqueTokens=new HashMap<Token, Token>();
    s = prepare(s);
    Token currentTitle = Token.IGNORE;
    List<Token> result = new ArrayList<Token>(s.length() / 5);
    
    // Start all pattern runners
    List<PatternRunner> activePatternRunners = new ArrayList<PatternRunner>(30);        
    for (PatternRunner p : patternRunners()) activePatternRunners.add(p);
    for (PatternRunner p : activePatternRunners) p.runOn(s);
    Announce.done();
    
    // Run through the String
    Announce.progressStart("Tokenizing",s.length());
    int position = 0;
    while (true) {
      Announce.progressAt(position);
      // Find best matching pattern
      PatternRunner best = null;
      for (int i = 0; i < activePatternRunners.size(); i++) {
        PatternRunner p = activePatternRunners.get(i);
        switch (p.beatsForPosition(best, position)) {
          case 1:
            best = p;
            break;
          case 0:
            activePatternRunners.remove(i);
            i--;
            continue;
        }
      }
      if (best == null) break;
      
      // Handle comments
      if (best.tokenType == Token.StartCommentToken.class) {
        Matcher m = ((Token.StartCommentToken) best.token).end.matcher(s);
        if (m.find(position)) position = m.end();
        else break;
        continue;
      }
      
      // Add the new Token      
      Token added=best.addReplacement(result, currentTitle,uniqueTokens);      
      position = best.newPosition();      
      if(added==null) continue;         
      
      // Recognize the title in the future
      if (added!=null && best.tokenType==Token.Title.class) {        
          currentTitle=added;
          String title = added.token();          
          Token.ProperName nameToken=new Token.ProperName(title);
          PatternRunner p=new PatternRunner(Pattern.quote(title), nameToken).runOn(s);
          for (int i = activePatternRunners.size()-1; i>=0; i--) {
            if(activePatternRunners.get(i).pattern().equals(p.pattern())) {
              activePatternRunners.remove(i);
              break;
            }
          }
          activePatternRunners.add(p); 
      }            
    }
    activePatternRunners=null; 
    Announce.progressDone();
    return (result);
  }
  
  /** Loads the file into memory*/
  public CharSequence load(File f) throws IOException {
    StringBuilder s = new StringBuilder((int) f.length());
    Reader in = new BufferedReader(new FileReader(f));
    int c;
    while ((c = in.read()) != -1) s.append((char) c);
    in.close();
    return (s);
  }

  /** Tokenizes a file*/
  public List<Token> tokenize(File f) throws IOException {
    return (tokenize(load(f)));
  }


  // ------------------------------------------------------------------------------------------
  //                       Tokenizer for HTML
  // ------------------------------------------------------------------------------------------

  /** Tokenizer for HTML*/
  public static class HTMLTokenizer extends Tokenizer {

    /** HTML-specific patterns*/
    protected static final List<PatternRunner> htmlPatternRunners=new ArrayList<PatternRunner>(standardPatternRunners);
    static {           
      htmlPatternRunners.add(new PatternRunner("</?[^>]+>", Token.IGNORE));
      htmlPatternRunners.add(new PatternRunner("</?LI>", Pattern.CASE_INSENSITIVE, Token.SEPARATOR));
      htmlPatternRunners.add(new PatternRunner("</?P>", Pattern.CASE_INSENSITIVE, Token.SEPARATOR));
      htmlPatternRunners.add(new PatternRunner("</?BR>", Pattern.CASE_INSENSITIVE, Token.SEPARATOR));
      htmlPatternRunners.add(new PatternRunner("<SCRIPT", Pattern.CASE_INSENSITIVE, Token.STARTSCRIPT));
      htmlPatternRunners.add(new PatternRunner("<!--", Token.HTMLCOMMENT));
    }

    @Override
    protected List<PatternRunner> patternRunners() {     
      return(htmlPatternRunners);
    }
    
    /** Decodes also Ampersands*/
    protected String prepare(CharSequence s) {
      return (super.prepare(Char.decodeAmpersand(s.toString())));
    }

    /** Loads the file, decodes UTF8*/
    public CharSequence load(File f) throws IOException {      
      StringBuilder s = new StringBuilder((int) f.length());
      Reader in = new UTF8Reader(f);
      int c;
      // We have to avoid very long sequences of characters that are not spaces
      // because else we get a Stackoverflow with the methods from java.util.regex.*...
      int numNonSpaces=0; 
      while ((c = in.read()) != -1)  {
        if(numNonSpaces<100) s.append((char) c);
        if(Character.isWhitespace(c)) numNonSpaces=0;
        else numNonSpaces++;
      }
      in.close();
      return (s);
    }
  }

  // ------------------------------------------------------------------------------------------
  //                       Tokenizer for Wikipedia
  // ------------------------------------------------------------------------------------------

  /** Tokenizer for Wikipedia*/
  public static class WikiTokenizer extends Tokenizer {

    /** Wikipedia-specific patterns*/
    protected static final List<PatternRunner> wikiPatternRunners=new ArrayList<PatternRunner>(standardPatternRunners);
    static {           
      wikiPatternRunners.add(new PatternRunner("</?[^>]++>", Token.IGNORE));
      wikiPatternRunners.add(new PatternRunner("''+", Token.IGNORE));
      wikiPatternRunners.add(new PatternRunner("[\\]\\[]++", Token.IGNORE));
      wikiPatternRunners.add(new PatternRunner("<!--", Token.HTMLCOMMENT));
      wikiPatternRunners.add(new PatternRunner("<ref", Token.REF));
      wikiPatternRunners.add(new PatternRunner("<id>\\d+</id>", Token.IGNORE));
      wikiPatternRunners.add(new PatternRunner("<revision>", Token.REVISION));
      wikiPatternRunners.add(new PatternRunner("<TITLE>([^<]++)</TITLE>", Pattern.CASE_INSENSITIVE, Token.Title.class));
      wikiPatternRunners.add(new PatternRunner("\\|", Token.INFOBOXBAR));
      wikiPatternRunners.add(new PatternRunner("\\[\\[([^\\]]++)\\]\\]", Token.WikiLink.class));
      wikiPatternRunners.add(new PatternRunner("\\[\\[([^\\|\\]]++)\\|[^\\]]++\\]\\]", Token.WikiLink.class));
      wikiPatternRunners.add(new PatternRunner("\\[\\[(\\p{Lu}[^\\]:]++)\\]\\]", Token.SemiCanonicProperName.class));
      wikiPatternRunners.add(new PatternRunner("\\[\\[(\\p{Lu}[^\\|\\]:]++\\|[^\\]]++)\\]\\]", Token.SemiCanonicProperName.class));
      wikiPatternRunners.add(new PatternRunner("\\[http:[^\\]]++\\]", Token.URL.class));
      wikiPatternRunners.add(new PatternRunner("\\[\\[Image:[^\\]]+\\]\\]", Token.IGNORE));
      wikiPatternRunners.add(new PatternRunner("==++([^=]++)==++", Token.class));      
      wikiPatternRunners.add(new PatternRunner("\\{\\{", Token.STARTBRACES));
      wikiPatternRunners.add(new PatternRunner("\\[\\[(\\d{4}-\\d{1,2}-\\d{1,2})\\]\\]", Token.Date.class));
      wikiPatternRunners.add(new PatternRunner("\\[\\[(\\d{4})\\]\\]", Token.Date.class));
      wikiPatternRunners.add(new PatternRunner("\\{\\{[\\w _]+\\|(\\d{4}-\\d{1,2}-\\d{1,2})\\}\\}", Token.Date.class));
      wikiPatternRunners.add(new PatternRunner("\\[\\[Category:",Token.CATEGORY));
      wikiPatternRunners.add(new PatternRunner("\\{\\{\\s*+inf", Pattern.CASE_INSENSITIVE, Token.INFOBOXHEAD));
      wikiPatternRunners.add(new PatternRunner("obox (\\p{L}++)", Pattern.CASE_INSENSITIVE, Token.ProperName.class));
      wikiPatternRunners.add(new PatternRunner("\\*", Token.SEPARATOR));
    }

    @Override
    protected List<PatternRunner> patternRunners() {     
      return(wikiPatternRunners);
    }

    /** Decodes Ampersand*/
    protected String prepare(CharSequence s) {
      return (super.prepare(Char.decodeAmpersand(s.toString())));
    }

    /** Loads the file, decodes UTF8*/
    public CharSequence load(File f) throws IOException {
      StringBuilder s = new StringBuilder((int) f.length());
      Reader in = new UTF8Reader(f);
      int c;
      while ((c = in.read()) != -1) s.append((char) c);
      in.close();
      return (s);
    }
        
    /** A pattern runner that eliminates all infoboxes*/
    protected static final PatternRunner infoboxKiller=new PatternRunner("\\{\\{\\s*+inf", Pattern.CASE_INSENSITIVE,Token.STARTINFOBOX);

    /** A pattern runner that eliminates everything except infoboxes*/
    protected static final PatternRunner infoboxSaver=new PatternRunner("'''", Token.STARTARTICLE);
    
    /** Tokenizes the Wikipedia article, but removes the infobox*/
    public List<Token> tokenizeNoInfobox(CharSequence s) {
      wikiPatternRunners.add(infoboxKiller);    
      List<Token> result=super.tokenize(s);
      wikiPatternRunners.remove(infoboxKiller);    
      return(result);
    }
    
    /** Tokenizes the Wikipedia article, but removes everything outside the infobox*/
    public List<Token> tokenizeOnlyInfobox(CharSequence s) {
      wikiPatternRunners.add(infoboxSaver);    
      List<Token> result=super.tokenize(s);
      wikiPatternRunners.remove(infoboxSaver);    
      return(result);
    }

    /** Tokenizes the Wikipedia article, but removes the infobox*/
    public List<Token> tokenizeNoInfobox(File f) throws IOException {
      return(tokenizeNoInfobox(load(f)));
    }
    
    /** Tokenizes the Wikipedia article, but only the infobox*/
    public List<Token> tokenizeOnlyInfobox(File f) throws IOException {
      return(tokenizeOnlyInfobox(load(f)));
    }
  }

  // ------------------------------------------------------------------------------------------
  //                       Main methods
  // ------------------------------------------------------------------------------------------

  /** Returns a tokenizer for a file*/
  public static Tokenizer forFile(File f) {
    if (FileSet.extension(f).toUpperCase().startsWith(".HTM")) return (new HTMLTokenizer());
    else if (FileSet.extension(f).toUpperCase().startsWith(".WIK")) return (new WikiTokenizer());
    else return (new Tokenizer());
  }

  /** Tokenizes a file*/
  public static List<Token> tokenizeString(String text) throws IOException {
    return new Tokenizer().tokenize(text);
  }
  
  /** Tokenizes a file*/
  public static List<Token> autodetectTokenize(File f) throws IOException {
    return(Tokenizer.forFile(f).tokenize(f));
  }

  /** Tokenizes a Wikipedia article, omitting the infoboxes*/
  public static List<Token> tokenizeWikipediaNoInfobox(File f) throws IOException {
    return(new WikiTokenizer().tokenizeNoInfobox(f));
  }

  /** Tokenizes the infobox of a Wikipedia article*/
  public static List<Token> tokenizeWikipediaOnlyInfobox(File f) throws IOException {
    return(new WikiTokenizer().tokenizeOnlyInfobox(f));
  }

  /** Tokenizes a Wikipedia article completely*/
  public static List<Token> tokenizeWikipedia(File f) throws IOException {
    return(new WikiTokenizer().tokenize(f));
  }

  /** Tokenizes an HTML document*/
  public static List<Token> tokenizeHTML(File f) throws IOException {
    return(new HTMLTokenizer().tokenize(f));
  }

  /** Tokenizes an text document*/
  public static List<Token> tokenizeText(File f) throws IOException {
    return(new Tokenizer().tokenize(f));
  }

  /** Test */
  public static void main(String[] args) throws Exception 
  {   
    D.println(autodetectTokenize(new File("c:/fabian/data/wikisub/companies/adobe systems.wiki")));
  }

}
